/* 
**  certmap.c.c -- Apache LDAP authorization module
**
**  Read the files README and mod_authz_ldap.html for instructions on
**  configuring the module. Details of the license can be found in the
**  HTML documentation.
**
**  (c) 2000 Dr. Andreas Mueller
**
**  $Id: certmap.c,v 1.3 2002/10/10 08:36:05 afm Exp $
*/ 
#include "mod_authz_ldap.h"
#ifdef EAPI
#include <pem.h>
extern char	*ssl_var_lookup(apr_pool_t *, server_rec *, conn_rec *,
			request_rec *, char *);
#ifndef AUTHZ_LDAP_HAVE_SSL
#define AUTHZ_LDAP_HAVE_SSL
#endif /* AUTHZ_LDAP_HAVE_SSL */
#endif /* EAPI */

static char *do_ssl_var_lookup(apr_pool_t *p, server_rec *s, conn_rec *c,
                               request_rec *r, char *v)
{
#ifdef EAPI
    return ssl_var_lookup(p, s, c, r, v);
#else
    APR_OPTIONAL_FN_TYPE(ssl_var_lookup) *lookupfn;

    lookupfn = APR_RETRIEVE_OPTIONAL_FN(ssl_var_lookup);
    
    if (lookupfn) {
        return lookupfn(p, s, c, r, v);
    } else {
        return NULL;
    }
#endif
}

/*************************************************************************
** Certificate Retrieval						**
*************************************************************************/

static char	*authz_ldap_get_issuer_dn(request_rec *r) {
#ifdef AUTHZ_LDAP_HAVE_SSL
	char			*issuer_dn;
	authz_ldap_config_rec	*sec;

	sec = ap_get_module_config(r->per_dir_config, &authz_ldap_module);
	issuer_dn = do_ssl_var_lookup(r->pool, r->server, r->connection, r,
		"SSL_CLIENT_I_DN");
	AUTHZ_DEBUG3("[%d] client issuer DN: %s", (int)getpid(),
		(issuer_dn) ? issuer_dn : "(null)");
	return issuer_dn;
#else /* AUTHZ_LDAP_HAVE_SSL */
	return NULL;
#endif /* AUTHZ_LDAP_HAVE_SSL */
}

static char	*authz_ldap_get_certificate(request_rec *r) {
#ifdef AUTHZ_LDAP_HAVE_SSL
	BIO		*bio_mem;
	int		l, j;
	unsigned char	*v;
	char		*xp, *cp, *p;
	authz_ldap_config_rec	*sec;
	X509		*xs;

	sec = ap_get_module_config(r->per_dir_config, &authz_ldap_module);

	/* retrieve the certificate in PEM form				*/
	xp = do_ssl_var_lookup(r->pool, r->server, r->connection, r,
		"SSL_CLIENT_CERT");
	if (NULL == xp) { 
		AUTHZ_DEBUG2("[%d] not certificate found", (int)getpid());
		return NULL;
	}

	/* turn the PEM data into a X.509 structure			*/
	bio_mem = BIO_new(BIO_s_mem());
	BIO_write(bio_mem, xp, strlen(xp) + 1);
	xs = PEM_read_bio_X509(bio_mem, NULL, NULL, NULL);
	BIO_free(bio_mem);
	if (NULL == xs) {
		AUTHZ_DEBUG3("[%d] could not read certificate from PEM: %s",
			(int)getpid(), xp);
		return NULL;
	}

	/* write the data from the certificate into DER form		*/
	bio_mem = BIO_new(BIO_s_mem());
	i2d_X509_bio(bio_mem, xs);
	l = BIO_ctrl_pending(bio_mem);
	v = ap_palloc(r->pool, l);
	BIO_read(bio_mem, v, l);
	BIO_free(bio_mem);

	/* quote the whole stuff so that it can safely be used in a 	*/
	/* filter							*/
	cp = ap_palloc(r->pool, 3 * l + 1);
	p = cp;
	for (j = 0; j < l; j++) {
		ap_snprintf(p, 4, "\\%02x", (unsigned char)v[j]);
		p += 3;
	}

	/* return the quoted certificate				*/
	return cp;
#else /* AUTHZ_LDAP_HAVE_SSL */
	return NULL;
#endif /* AUTHZ_LDAP_HAVE_SSL */
}

static char	*authz_ldap_get_subject_dn(request_rec *r) {
#ifdef AUTHZ_LDAP_HAVE_SSL
	char	*subject_dn;
	authz_ldap_config_rec	*sec;

	sec = ap_get_module_config(r->per_dir_config, &authz_ldap_module);
	subject_dn = do_ssl_var_lookup(r->pool, r->server, r->connection, r,
		"SSL_CLIENT_S_DN");
	AUTHZ_DEBUG3("[%d] client subject DN: %s", (int)getpid(),
		(subject_dn) ? subject_dn : "(null)");
	return subject_dn;
#else /* AUTHZ_LDAP_HAVE_SSL */
	return NULL;
#endif /* AUTHZ_LDAP_HAVE_SSL */
}

static char	*authz_ldap_get_serial(request_rec *r) {
#ifdef AUTHZ_LDAP_HAVE_SSL
	char	*cp;
	authz_ldap_config_rec	*sec;

	sec = ap_get_module_config(r->per_dir_config, &authz_ldap_module);

	/* get serial number from X509 client certificate		*/
	cp = do_ssl_var_lookup(r->pool, r->server, r->connection, r,
		"SSL_CLIENT_M_SERIAL");
	AUTHZ_DEBUG3("[%d] client serial: %s", (int)getpid(),
		(cp) ? cp : "(null)");
	return cp;
#else /* AUTHZ_LDAP_HAVE_SSL */
	return NULL;
#endif /* AUTHZ_LDAP_HAVE_SSL */
}

/*************************************************************************
** Certificate Mapping							**
*************************************************************************/

/* utility function: map a certificate to a user name			*/
/* returns 1 if mapping was successful, 0 otherwise			*/
int	authz_ldap_map_user(request_rec *r) {
	authz_ldap_config_rec	*sec;
	char			dn[MAX_STRING_LEN];
	char			*attr, *issuer, *subject, *serial,
				*certificate;
	const char		*userdn;
	char			*base;
	int			scope;
	int			nentries;
	LDAPMessage		*result, *e;
	char			filter[MAX_STRING_LEN];
	char			*attrs[2];
	BerElement		*ber;
	char			**vals;

	sec = ap_get_module_config(r->per_dir_config, &authz_ldap_module);

	/* if the connection notes table contains a value for 		*/
	/* authz_ldap::userdn, the mapping for the certificate was	*/
	/* successful, and we don't need to map again.			*/
	userdn = ap_table_get(r->notes, "authz_ldap::userdn");
	if ((userdn != NULL) && (strlen(userdn) > 0)) {
		AUTHZ_DEBUG3("[%d] user mapping known: %s", (int)getpid(),
			userdn);
		return 1;
	}

	/* find issuer and subject distinguished name (short form)	*/
	issuer = authz_ldap_get_issuer_dn(r);
	if (NULL == issuer) {
		AUTHZ_DEBUG2("[%d] issuer not found in certificate",
			(int)getpid());
		return 0;
	}
	subject = authz_ldap_get_subject_dn(r);
	if (NULL == subject) {
		AUTHZ_DEBUG2("[%d] subject not found in certificate",
			(int)getpid());
		return 0;
	}
	serial = authz_ldap_get_serial(r);
	if (NULL == serial) {
		AUTHZ_DEBUG2("[%d] serial not found in certificate",
			(int)getpid());
		return 0;
	}
	certificate = authz_ldap_get_certificate(r);
	if (NULL == certificate) {
		AUTHZ_DEBUG2("[%d] certificate not found in certificate",
			(int)getpid());
		return 0;
	}

	/* build a filter for the certificate map			*/
	if (sec->certdirect) {
		ap_snprintf(filter, MAX_STRING_LEN,
			"(&(userCertificate=%s)"
			"(objectClass=strongAuthenticationUser))", certificate);
		base = sec->userbase;
		scope = sec->userscope;
	} else if (sec->useserial) {
		ap_snprintf(filter, MAX_STRING_LEN,
			"(&(issuerDN=%s)(serialNumber=%s)"
			"(objectClass=authzLDAPmap))", issuer, serial);
		base = sec->mapbase;
		scope = sec->mapscope;
	} else {
		ap_snprintf(filter, MAX_STRING_LEN,
			"(&(issuerDN=%s)(subjectDN=%s)"
			"(objectClass=authzLDAPmap))", issuer, subject);
		base = sec->mapbase;
		scope = sec->mapscope;
	}
	AUTHZ_DEBUG4("[%d] authz_ldap_search for certificate mapping: "
		"base = %s, filter = %s", (int)getpid(), sec->mapbase, filter);

	/* we have to retrieve the dn from the directory		*/
	if (!sec->certdirect) {
		attrs[0] = "owner";
	} else {
		attrs[0] = NULL;
	}
	attrs[1] = NULL;
	if (authz_ldap_search(r, base, scope,
		filter, attrs, 0, &result) != LDAP_SUCCESS) {
		if (sec->loglevel >= APLOG_ERR)
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR,
			APLOG_STATUS r,
			"[%d] certificate owner for [ %s | %s ] not found",
			(int)getpid(), issuer, subject);
		return 0;
	}

	/* get the distinguished name from the directory		*/
	nentries = ldap_count_entries(sec->ldap, result);
	if (nentries == 0) {
		ldap_msgfree(result);
		if (sec->loglevel >= APLOG_ERR)
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR,
			APLOG_STATUS r,
			"[%d] filter: %s base: %s, no such user", (int)getpid(),
			filter, base);
		return 0;
	}
	if (nentries > 1) {
		ldap_msgfree(result);
		if (sec->loglevel >= APLOG_ERR)
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR,
			APLOG_STATUS r,
			"[%d] filter: %s base: %s, not unique", (int)getpid(),
			filter, base);
		return 0;
	}

	/* get the attribute value for the owner attribute		*/
	dn[0] = '\0';
	e = ldap_first_entry(sec->ldap, result);
	if (e != NULL) {
		if (sec->certdirect) {
			char	*d;
			d = ldap_get_dn(sec->ldap, e);
			apr_cpystrn(dn, d, MAX_STRING_LEN);
			ldap_memfree(d);
		} else {
			attr = ldap_first_attribute(sec->ldap, e, &ber);
			if (attr != NULL) {
				vals = ldap_get_values(sec->ldap, e, attr);
				if (vals != NULL) {
					apr_cpystrn(dn, vals[0], MAX_STRING_LEN);
					ldap_value_free(vals);
				}
				if (ber != NULL)
					ber_free(ber, 0);
			}
		}
	}
	ldap_msgfree(result);

	AUTHZ_DEBUG3("[%d] owner dn: %s", (int)getpid(), dn);

	/* remember the distinguished named of the user			*/
	authz_ldap_set_userdn(r, dn);

	/* set the the user to the distinguished name of the		*/
	/* certificate owner						*/
	AUTHZ_DEBUG5("[%d] user [ %s | %s ] mapped to '%s'", (int)getpid(),
		issuer, subject, dn);

	/* remember the distinguished name of the user in the notes	*/
	userdn = ap_pstrdup(r->pool, dn);
	AUTHZ_DEBUG3("[%d] setting authz_ldap::userdn notes table entry "
		"to '%s'", (int)getpid(), userdn);
	ap_table_set(r->notes, "authz_ldap::userdn", userdn);

	/* overwrite the authentication header if necessary		*/
	if (sec->setauth)
		authz_ldap_set_user_to_dn(r, userdn);

	/* that's it, if we get to this point, we can assume the 	*/
	/* mapping was a success					*/
	return 1;
}
