/* 
**  modconfig.c.c -- Apache LDAP authorization module configuration
**
**  Read the files README and mod_authz_ldap.html for instructions on
**  configuring the module. Details of the license can be found in the
**  HTML documentation.
**
**  (c) 2001 Dr. Andreas Mueller
**
**  $Id: modconf.c,v 1.6 2002/10/10 08:36:05 afm Exp $
*/ 
#include "mod_authz_ldap.h"

#ifndef AP_SERVER_MAJORVERSION
/* The content handler */
/* XXX it is planned that this displays statistics about the module	*/
static int authz_ldap_handler(request_rec *r)
{
    r->content_type = "text/html";      
    ap_send_http_header(r);
    if (!r->header_only)
        ap_rputs("Authz_LDAP statistics are not yet implemented.\n", r);
    return OK;
}
#endif /* AP_SERVER_MAJORVERSION */

#ifndef AP_SERVER_MAJORVERSION
/* Dispatch list of content handlers */
const handler_rec authz_ldap_handlers[] = { 
    { "authz_ldap", authz_ldap_handler }, 
    { NULL, NULL }
};
#endif /* AP_SERVER_MAJORVERSION */

/*************************************************************************
** Configuration							**
*************************************************************************/

/* parse scope configuration directive					*/
const char	*authz_ldap_set_user_scope_slot(cmd_parms *parms,
	void *mconfig, const char *arg) {
	authz_ldap_config_rec	*sec;
	sec = (authz_ldap_config_rec *)mconfig;

#if AUTHZ_LDAP_DEBUG
	fprintf(stderr, "%s:%d: setting user search scope at %p to '%s'\n",
		__FILE__, __LINE__, (void *)sec, (arg) ? arg : "(null)");
#endif
	/* check for correct directives					*/
	if (0 == strcasecmp("subtree", arg)) {
		sec->userscope = LDAP_SCOPE_SUBTREE;
	} else if (0 == strcasecmp("onelevel", arg)) {
		sec->userscope = LDAP_SCOPE_ONELEVEL;
	} else if (0 == strcasecmp("base", arg)) {
		sec->userscope = LDAP_SCOPE_BASE;
	} else {
		return "illegal argument to AuthzLDAPUserScope";
	}
	return NULL;
}

const char	*authz_ldap_set_group_scope_slot(cmd_parms *parms,
	void *mconfig, const char *arg) {
	authz_ldap_config_rec	*sec;
	sec = (authz_ldap_config_rec *)mconfig;

#if AUTHZ_LDAP_DEBUG
	fprintf(stderr, "%s:%d: setting group search scope at %p to '%s'\n",
		__FILE__, __LINE__, (void *)sec, (arg) ? arg : "(null)");
#endif
	/* check for correct directives					*/
	if (0 == strcasecmp("subtree", arg)) {
		sec->groupscope = LDAP_SCOPE_SUBTREE;
	} else if (0 == strcasecmp("onelevel", arg)) {
		sec->groupscope = LDAP_SCOPE_ONELEVEL;
	} else if (0 == strcasecmp("base", arg)) {
		sec->groupscope = LDAP_SCOPE_BASE;
	} else {
		return "illegal argument to AuthzLDAPUserScope";
	}
	return NULL;
}

const char	*authz_ldap_set_map_scope_slot(cmd_parms *parms,
	void *mconfig, const char *arg) {
	authz_ldap_config_rec	*sec;
	sec = (authz_ldap_config_rec *)mconfig;

#if AUTHZ_LDAP_DEBUG
	fprintf(stderr, "%s:%d: setting map search scope at %p to '%s'\n",
		__FILE__, __LINE__, (void *)sec, (arg) ? arg : "(null)");
#endif
	/* check for correct directives					*/
	if (0 == strcasecmp("subtree", arg)) {
		sec->mapscope = LDAP_SCOPE_SUBTREE;
	} else if (0 == strcasecmp("onelevel", arg)) {
		sec->mapscope = LDAP_SCOPE_ONELEVEL;
	} else if (0 == strcasecmp("base", arg)) {
		sec->mapscope = LDAP_SCOPE_BASE;
	} else {
		return "illegal argument to AuthzLDAPMapScope";
	}
	return NULL;
}

const char	*authz_ldap_set_loglevel_slot(cmd_parms *parms,
	void *mconfig, const char *arg) {
	authz_ldap_config_rec	*sec;
	sec = (authz_ldap_config_rec *)mconfig;

#if AUTHZ_LDAP_DEBUG
	fprintf(stderr, "%s:%d: setting loglevel at %p to '%s'\n",
		__FILE__, __LINE__, (void *)sec, (arg) ? arg : "(null)");
#endif
	/* check for the various possible legal values of the argument	*/
	if (0 == strcasecmp("emerg", arg)) {
		sec->loglevel = APLOG_EMERG; return NULL;
	}
	if (0 == strcasecmp("alert", arg)) {
		sec->loglevel = APLOG_ALERT; return NULL;
	}
	if (0 == strcasecmp("crit", arg)) {
		sec->loglevel = APLOG_CRIT; return NULL;
	}
	if (0 == strcasecmp("error", arg)) {
		sec->loglevel = APLOG_ERR; return NULL;
	}
	if (0 == strcasecmp("warn", arg)) {
		sec->loglevel = APLOG_WARNING; return NULL;
	}
	if (0 == strcasecmp("notice", arg)) {
		sec->loglevel = APLOG_NOTICE; return NULL;
	}
	if (0 == strcasecmp("info", arg)) {
		sec->loglevel = APLOG_INFO; return NULL;
	}
	if (0 == strcasecmp("debug", arg)) {
		sec->loglevel = APLOG_DEBUG; return NULL;
	}
	return "illegal argument to AuthzLDAPLogLevel";
}

#ifdef HAVE_LDAP_ENABLE_CACHE
/*
 * This part of the code implements the configuration parameters for
 * the LDAP cache, and is only compiled into the code if the library
 * implements the ldap_enable_cache function.
 */

/* read the cache timeout from the configuration			*/
const char	*authz_ldap_set_timeout_slot(cmd_parms *parms,
	void *mconfig, const char *arg) {
	authz_ldap_config_rec	*sec;
	sec = (authz_ldap_config_rec *)mconfig;

	sec->timeout = atoi(arg);
	if ((sec->timeout < 0) || (sec->timeout > 86400)) {
		ap_log_error(APLOG_MARK, APLOG_ERR, APLOG_STATUS parms->server,
			"[%d] illegal LDAP cache timeout: %d, using default %d",
			(int)getpid(), sec->timeout,
			AUTHZ_DEFAULT_CACHE_TIMEOUT);
		sec->timeout = AUTHZ_DEFAULT_CACHE_TIMEOUT;
	}
	return NULL;
}

/* read the cache size from the configuration				*/
const char	*authz_ldap_set_cachesize_slot(cmd_parms *parms,
	void *mconfig, const char *arg) {
	authz_ldap_config_rec	*sec;
	sec = (authz_ldap_config_rec *)mconfig;

	sec->cachesize = atoi(arg);
	if ((sec->cachesize < 0) || (sec->cachesize > AUTHZ_MAX_CACHE_SIZE)) {
		ap_log_error(APLOG_MARK, APLOG_ERR, APLOG_STATUS parms->server,
			"[%d] illegal LDAP cache size: %d, cache disabled",
			(int)getpid(), sec->cachesize);
		sec->cachesize = 0;
	}
	return NULL;
}

#endif /* HAVE_LDAP_ENABLE_CACHE */

const char	*authz_ldap_set_version_slot(cmd_parms *parms,
	void *mconfig, const char *arg) {
	authz_ldap_config_rec	*sec;
	sec = (authz_ldap_config_rec *)mconfig;

	sec->ldapversion = atoi(arg);
	if (0 == sec->ldapversion) {
		ap_log_error(APLOG_MARK, APLOG_WARNING,
			APLOG_STATUS parms->server,
			"[%d] version cannot be set: %s", (int)getpid(),
			arg);
	}
	return NULL;
}

/* Configuration Directives for this module				*/
const command_rec	authz_ldap_cmds[] = {
	/* whether or not to be active at all!				*/
	AP_INIT_FLAG(
		"AuthzLDAPEngine", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, engine),
		OR_AUTHCFG,
		"Set to 'on' if mod_authz_ldap should become active"
	),
	AP_INIT_FLAG(
		"AuthzLDAPCacheConnection", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, cacheconnection),
		OR_AUTHCFG,
		"Set to 'on' if mod_authz_ldap should cache the connection to the directory"
	),
	/* User settings						*/
	AP_INIT_TAKE1(
		"AuthzLDAPServer", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, server),
		OR_AUTHCFG,
		"Name of LDAP Server that should be queried"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPBindDN", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, binddn),
		OR_AUTHCFG,
		"DN as which to bind to the LDAP directory"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPBindPassword", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, bindpw),
		OR_AUTHCFG,
		"The password to use when binding to the directory"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPProtocolVersion", authz_ldap_set_version_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, ldapversion),
		OR_AUTHCFG,
		"The version of the LDAP protocol to use to connect to the directory"
	),

	/* User settings						*/
	AP_INIT_TAKE1(
		"AuthzLDAPUserKey", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, userkey),
		OR_AUTHCFG,
		"The attribute name associated with the userid"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPUserBase", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, userbase),
		OR_AUTHCFG,
		"The DN of the node immediate above the users"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPUserScope", authz_ldap_set_user_scope_slot, NULL,
		OR_AUTHCFG,
		"the scope for a search for the user: subtree, onelevel, base"
	),
	

	/* Group membership settings					*/
	AP_INIT_TAKE1(
		"AuthzLDAPGroupBase", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, groupbase),
		OR_AUTHCFG,
		"Base DN for searches for groups"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPGroupKey", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, groupkey),
		OR_AUTHCFG,
		"attribute name for groups"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPGroupScope", authz_ldap_set_group_scope_slot, NULL,
		OR_AUTHCFG,
		"the scope for a search for the group: subtree, onelevel, base"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPMemberKey", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, memberkey),
		OR_AUTHCFG,
		"attribute name for group members"
	),

	/* X.509 certificate mapping					*/
	AP_INIT_FLAG(
		"AuthzLDAPDirect", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, certdirect),
		OR_AUTHCFG,
		"Set to 'on' if authentication should try to search for the certificate directly"
	),
	AP_INIT_FLAG(
		"AuthzLDAPUseCertificate", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, usecertificate),
		OR_AUTHCFG,
		"Set to 'on' if user should be authenticated from SSL alone"
	),
	AP_INIT_FLAG(
		"AuthzLDAPUseSerial", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, useserial),
		OR_AUTHCFG,
		"Set to 'on' if user mapping should be based on certificate serial number"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPMapBase", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, mapbase),
		OR_AUTHCFG,
		"Where to start searching during a certificate lookup"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPMapScope", authz_ldap_set_map_scope_slot, NULL,
		OR_AUTHCFG,
		"the scope for a search during the mapping of a X.509 "
			"certificate: subtree, onelevel, base"
	),
	AP_INIT_FLAG(
		"AuthzLDAPSetAuthorization", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, setauth),
		OR_AUTHCFG,
		"Set to 'off' if user DN should not appear in authorization header, default is 'on'"
	),

	/* use LDAP flag for double authentication			*/
	AP_INIT_FLAG(
		"AuthzLDAPUseLDAP", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, useldap),
		OR_AUTHCFG,
		"Set to 'on' if user should be authenticated from SSL alone and LDAP"
	),

	/* name of attribute containing role names			*/
	AP_INIT_TAKE1(
		"AuthzLDAPRoleAttributeName", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, roleattrname),
		OR_AUTHCFG,
		"Name of attribute containing role designations"
	),

	/* Password aging mechanisms					*/
	AP_INIT_TAKE1(
		"AuthzLDAPModifyKey", ap_set_string_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, modifykey),
		OR_AUTHCFG,
		"Key to look for when trying to check for the last modification"
	),

#ifdef HAVE_LDAP_ENABLE_CACHE
	/* LDAP cache size						*/
	AP_INIT_TAKE1(
		"AuthzLDAPCacheTimeout", authz_ldap_set_timeout_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, timeout),
		OR_AUTHCFG,
		"Timout for entries in the LDAP cache in seconds, default 600, i.e. 10 minutes"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPCacheSize", authz_ldap_set_cachesize_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, cachesize),
		OR_AUTHCFG,
		"Size of the LDAP cache, default 0, i.e. no caching"
	),
#endif /* HAVE_LDAP_ENABLE_CACHE */

	/* general variables						*/
	AP_INIT_FLAG(
		"AuthzLDAPAuthoritative", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, authoritative),
		OR_AUTHCFG,
		"Set to 'off' if you want to hand authentication to some "
			"other authentication handler further down"
	),
	AP_INIT_FLAG(
		"AuthzLDAPProxyAuthentication", ap_set_flag_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, proxyauth),
		OR_AUTHCFG,
		"Set to 'on' if proxy authentication is happening here"
	),
	AP_INIT_TAKE1(
		"AuthzLDAPLogLevel", authz_ldap_set_loglevel_slot,
		(void *)APR_OFFSETOF(authz_ldap_config_rec, loglevel),
		OR_AUTHCFG,
		"limit the debug information from the module"
	),

	{ NULL }
};

/* initialize a directory configuration record				*/
void	*authz_ldap_create_dir_config(apr_pool_t *p, char *d) {
	authz_ldap_config_rec	*sec
		= (authz_ldap_config_rec *)ap_palloc(p,
			sizeof(authz_ldap_config_rec));
#if AUTHZ_LDAP_DEBUG
	fprintf(stderr, "%s:%d: initializing dir config record at %p\n",
		__FILE__, __LINE__, (void *)sec);
#endif
	if (sec) {
		/* don't be active/caching connections by default	*/
		sec->engine = 0;
		sec->cacheconnection = 0;

		/* global configuration information			*/
		sec->server = NULL;
		sec->binddn = NULL;
		sec->bindpw = NULL;
#ifdef LDAP_VERSION3
		sec->ldapversion = LDAP_VERSION3;
#else
		sec->ldapversion = 0;	/* by default, don't set the	*/
					/* protocol version		*/
#endif

		/* user configuration					*/
		sec->userbase = NULL;
		sec->userkey = NULL;
		sec->userscope = LDAP_SCOPE_BASE;

		/* group configuration					*/
		sec->groupbase = NULL;
		sec->groupkey = NULL;
		sec->groupscope = LDAP_SCOPE_BASE;
		sec->memberkey = NULL;

		/* X.509 certificate mapping				*/
		sec->certdirect = 0;
		sec->usecertificate = 0;
		sec->useserial = 0;
		sec->mapbase = NULL;
		sec->mapscope = LDAP_SCOPE_SUBTREE;
		sec->setauth = 1;

		/* additional LDAP authentication			*/
		sec->useldap = 0; 	/* don't use the DN unless told	*/

		/* role attribute name					*/
		sec->roleattrname = NULL;

		/* password aging					*/
		sec->modifykey = NULL;

#ifdef HAVE_LDAP_ENABLE_CACHE
		/* LDAP cache						*/
		sec->timeout = AUTHZ_DEFAULT_CACHE_TIMEOUT;
		sec->cachesize = 0;	/* disabled by default		*/
#endif /* HAVE_LDAP_ENABLE_CACHE */

		/* generic configuration				*/
		sec->authoritative = 1;
		sec->proxyauth = AUTHZ_AUTO;
		sec->loglevel = APLOG_DEBUG;

		/* private data						*/
		sec->ldap = NULL;
		sec->username = NULL;
		sec->userdn = NULL;
	}
	return sec;
}

/* directory configuration merger					*/
#define	authz_ldap_link(f)	do {					\
		if ((parent->f) && (!child->f)) child->f = parent->f;	\
	} while (0)
void	*authz_ldap_merge_dir_config(apr_pool_t *p, void *pp, void *cp) {
	authz_ldap_config_rec	*parent = (authz_ldap_config_rec *)pp;
	authz_ldap_config_rec	*child = (authz_ldap_config_rec *)cp;
	
	/* all values that were not set in the child but are set in the	*/
	/* parent are linked						*/
	authz_ldap_link(server);
	authz_ldap_link(binddn);
	authz_ldap_link(bindpw);

	authz_ldap_link(userbase);
	authz_ldap_link(userkey);

	authz_ldap_link(groupbase);
	authz_ldap_link(groupkey);
	authz_ldap_link(memberkey);
	
	authz_ldap_link(mapbase);

	/* connections possibly cached from previous calls should be	*/
	/* restored							*/
	authz_ldap_link(ldap);
	
	/* the integers cannot possibly be merged, as we cannot decide	*/
	/* from the values in the child structure whether they are 	*/
	/* defaults, or set by the parent				*/
	return child;
}

